#!/usr/bin/env python
""" Tolk

Usage:
    tolk [--socket=<path> --modbus-host=<host> --modbus-port=<nr>]

Options:
    -h --help           Show this screen.
    --socket=<path>     Location of Tolk's socket [default: /tmp/tolk.sock].
    --modbus-host=<ip>  IP of Modbus slave [default: localhost].
    --modbus-port=<nr>  Port of Modbus slave [default: 502]

"""
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                '../'))

from logbook import Logger, StreamHandler
from modbus_tk.modbus_tcp import TcpMaster
from SocketServer import UnixStreamServer
from docopt import docopt

from tolk import Dispatcher, Handler

StreamHandler(sys.stdout).push_application()
log = Logger(__name__)


def main():
    args = docopt(__doc__)

    modbus_master = TcpMaster(args['--modbus-host'],
                              int(args['--modbus-port']))
    dispatcher = Dispatcher(modbus_master)

    server = UnixStreamServer(args['--socket'], Handler)
    server.dispatcher = dispatcher

    try:
        log.info('Start Tolk listening at {0}.'.format(args['--socket']))
        server.serve_forever()
    except KeyboardInterrupt:
        log.info('Received SIGINT. Exiting')
        pass
    finally:
        os.unlink(args['--socket'])
        log.info('Tolk has stopped')


if __name__ == '__main__':
    main()



from functools import wraps

from modbus_tk.modbus import ModbusError
from pyjsonrpc.rpcerror import jsonrpcerrors, JsonRpcError

modbus_mapping = {}


def json_rpc_error(func):
    """ Wraps a function and raises a JSON RPC error when a ModbusError has
    been excepted::

        @json_rpc_error()
        def modbus_request(modbus_master, slave_id, function_code,
                           starting_address, quantity):

        modbus_master.execute(int(slave_id), function_code,
                              int(starting_address),
                              int(quantity))

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ModbusError as e:
            raise modbus_mapping[e.get_exception_code()]
    return wrapper


class IllegalFunction(JsonRpcError):
    code = -32001
    message = 'Function code is not valid.'


jsonrpcerrors[IllegalFunction.code] = IllegalFunction
modbus_mapping[1] = IllegalFunction


class IllegalDataAddress(JsonRpcError):
    code = -32002
    message = 'Data address is not valid.'


jsonrpcerrors[IllegalDataAddress.code] = IllegalDataAddress
modbus_mapping[2] = IllegalDataAddress


class IllegalDataValue(JsonRpcError):
    code = -32003
    message = 'Data value is not valid.'


jsonrpcerrors[IllegalDataValue.code] = IllegalDataValue
modbus_mapping[3] = IllegalDataValue


class SlaveDeviceFailure(JsonRpcError):
    code = -32004
    message = 'Slave device could not perform requested action.'


jsonrpcerrors[SlaveDeviceFailure.code] = SlaveDeviceFailure
modbus_mapping[4] = SlaveDeviceFailure


class CommandAcknowledge(JsonRpcError):
    code = -32005
    message = 'Slave device has accepted the request and is ' \
              'processing it, but a long duration of time will be required to ' \
              'do so. This response is returned to prevent a timeout error ' \
              'from occurring in the master.'


jsonrpcerrors[CommandAcknowledge.code] = CommandAcknowledge
modbus_mapping[5] = CommandAcknowledge


class SlaveDeviceBusy(JsonRpcError):
    code = -32006
    message = 'Slave device is busy.'


jsonrpcerrors[SlaveDeviceBusy.code] = SlaveDeviceBusy
modbus_mapping[6] = SlaveDeviceBusy


class NegativeAcknowlegde(JsonRpcError):
    code = -32007
    message = 'Slave device cannot perform the program function received in the ' \
              'query.'


jsonrpcerrors[NegativeAcknowlegde.code] = NegativeAcknowlegde
modbus_mapping[7] = NegativeAcknowlegde


class MemoryParityError(JsonRpcError):
    code = -32008
    message = 'Slave device failed to read extended memory or record file.'


jsonrpcerrors[MemoryParityError.code] = MemoryParityError
modbus_mapping[8] = MemoryParityError


class GateWayPathUnavailable(JsonRpcError):
    code = -32010
    message = 'Gateway is unable to allocate an internal communication path from ' \
              'the input port to the output port.'


jsonrpcerrors[GateWayPathUnavailable.code] = GateWayPathUnavailable
modbus_mapping[10] = GateWayPathUnavailable


class GatewayTargetDeviceFailedToRespond(JsonRpcError):
    code = -32011
    message = 'No response obtained from the target device.'


jsonrpcerrors[GatewayTargetDeviceFailedToRespond.code] = GatewayTargetDeviceFailedToRespond
modbus_mapping[11] = GatewayTargetDeviceFailedToRespond



from modbus_tk.defines import (READ_COILS, READ_DISCRETE_INPUTS,
                               READ_HOLDING_REGISTERS, READ_INPUT_REGISTERS,
                               WRITE_SINGLE_COIL, WRITE_SINGLE_REGISTER,
                               WRITE_MULTIPLE_COILS, WRITE_MULTIPLE_REGISTERS)
from pyjsonrpc import JsonRpc, rpcmethod
from tolk.exceptions import json_rpc_error


class Dispatcher(JsonRpc):
    def __init__(self, modbus_master):
        self.modbus_master = modbus_master

        super(JsonRpc, self).__init__()

    @rpcmethod
    @json_rpc_error
    def read_coils(self, starting_address, quantity, slave_id=1):
        """ Execute Modbus function code 01: read status of coils.

        :param starting_address: Number of starting address.
        :param quantity: Number of coils to read.
        :param slave_id: Number with Slave id, default 1.
        :returns: JSON-RPC response containing the status of coils on success JSON-RPC error on failure.

        **Example request:**

        .. sourcecode:: json

            {
                "params":{
                    "starting_address":100,
                    "quantity":2,
                    "slave_id":1
                },
                "jsonrpc":"2.0",
                "method":"read_coils",
                "id":1
            }

        **Example response:**

        .. sourcecode:: json

            {
                "jsonrpc":"2.0",
                "id":1,
                "result":[
                    0,
                    1
                ]
            }
        """
        return self.modbus_master.execute(int(slave_id), READ_COILS,
                                          int(starting_address),
                                          int(quantity))

    @rpcmethod
    @json_rpc_error
    def read_discrete_inputs(self, starting_address, quantity, slave_id=1):
        """ Execute Modbus function code 02: read status of discrete inputs.

        :param starting_address: Number of starting address.
        :param quantity: Number of discrete inputs to read.
        :param slave_id: Number with Slave id, default 1.
        :returns: JSON-RPC response containing the status of discrete inputs
        on success or JSON-RPC error on failure.

        **Example request:**

        .. sourcecode:: json

            {
                "params":{
                    "starting_address":0,
                    "quantity":2,
                    "slave_id":1
                },
                "jsonrpc":"2.0",
                "method":"read_discrete_inputs",
                "id":1
            }

        **Example response:**

        .. sourcecode:: json

            {
                "jsonrpc":"2.0",
                "id":1,
                "result":[
                    0,
                    1,
                ]
            }
        """
        return self.modbus_master.execute(int(slave_id), READ_DISCRETE_INPUTS,
                                          int(starting_address),
                                          int(quantity))

    @rpcmethod
    @json_rpc_error
    def read_holding_registers(self, starting_address, quantity, slave_id=1):
        """ Execute Modbus function code 03: read contents of contiguous block
        of holding registers.

        :param starting_address: Number of starting address.
        :param quantity: Number of holding registers to read.
        :param slave_id: Number with Slave id, default 1.
        :returns: JSON-RPC response with the contents of holding registers
        on success or JSON-RPC error on failure.

        **Example request:**

        .. sourcecode:: json

            {
                "params":{
                    "starting_address":100,
                    "quantity":2,
                    "slave_id":1
                },
                "jsonrpc":"2.0",
                "method":"read_holding_registers",
                "id":1
            }

        **Example response:**

        .. sourcecode:: json

            {
                "jsonrpc":"2.0",
                "id":1,
                "result":[
                    1234,
                    32433,
                ]
            }
        """
        return self.modbus_master.execute(int(slave_id), READ_HOLDING_REGISTERS,
                                          int(starting_address),
                                          int(quantity))

    @rpcmethod
    @json_rpc_error
    def read_input_registers(self, starting_address, quantity, slave_id=1):
        """ Execute Modbus function code 04: read contents of contiguous block
        of input registers.

        :param starting_address: Number of starting address.
        :param quantity: Number of input registers to read.
        :param slave_id: Number with Slave id, default 1.
        :returns: JSON-RPC response with the contents of input registers
        on success or JSON-RPC error on failure.

        **Example request:**

        .. sourcecode:: json

            {
                "params":{
                    "starting_address":100,
                    "quantity":2,
                    "slave_id":1
                },
                "jsonrpc":"2.0",
                "method":"read_input_registers",
                "id":1
            }

        **Example response:**

        .. sourcecode:: json

            {
                "jsonrpc":"2.0",
                "id":1,
                "result":[
                    1234,
                    32433,
                ]
            }
        """
        return self.modbus_master.execute(int(slave_id), READ_INPUT_REGISTERS,
                                          int(starting_address),
                                          int(quantity))

    @rpcmethod
    @json_rpc_error
    def write_single_coil(self, address, value, slave_id=1):
        """ Execute Modbus function code 05: write value to single coil.

        :param address: Address of coil.
        :param value: Value to write to coil.
        :param slave_id: Number with Slave id, default 1.
        :returns: JSON-RPC response containing the address and value that has been written.

        **Example request:**

        .. sourcecode:: json

            {
                "params":{
                    "starting_address":100,
                    "value":1,
                    "slave_id":1
                },
                "jsonrpc":"2.0",
                "method":"write_single_coil",
                "id":1
            }

        **Example response:**

        .. sourcecode:: json

            {
                "jsonrpc":"2.0",
                "id":1,
                "result":[
                    100,
                    65280,
                ]
            }
        """
        return self.modbus_master.execute(int(slave_id), WRITE_SINGLE_COIL,
                                          int(address), output_value=int(value))

    @rpcmethod
    @json_rpc_error
    def write_single_register(self, address, value, slave_id=1):
        """ Execute Modbus function code 06: write value to single holding
        register.

        :param address: Address of holding register.
        :param value: Value to write to holding register.
        :param slave_id: Number with Slave id, default 1.
        :returns: JSON-RPC response containing the address and the value that has been written.

        **Example request:**

        .. sourcecode:: json

            {
                "params":{
                    "starting_address":100,
                    "value":1234,
                    "slave_id":1
                },
                "jsonrpc":"2.0",
                "method":"write_single_register",
                "id":1
            }

        **Example response:**

        .. sourcecode:: json

            {
                "jsonrpc":"2.0",
                "id":1,
                "result":[
                    100,
                    1234,
                ]
            }
        """
        return self.modbus_master.execute(int(slave_id),
                                          WRITE_SINGLE_REGISTER, int(address),
                                          output_value=int(value))

    @rpcmethod
    @json_rpc_error
    def write_multiple_coils(self, starting_address, values, slave_id=1):
        """ Execute Modbus function code 15: write sequence of values to a
        contiguous block of coils.

        .. note:: This method doesn't follow the interface as described in the
            Modbus specification. The specification describes 2 extra
            'parameters': quantity of outputs and byte count. Tolk will derive
            them from parameter `values`.

        :param starting_address: Number of starting address.
        :param values: List with values.
        :param slave_id: Number with Slave id, default 1.
        :returns: JSON-RPC response containing the address and the number of coils that has been written.

        **Example request:**

        .. sourcecode:: json

            {
                "params":{
                    "starting_address":100,
                    "value": [1, 0, 1, 0],
                    "slave_id":1
                },
                "jsonrpc":"2.0",
                "method":"write_multiple_coils",
                "id":1
            }

        **Example response:**

        .. sourcecode:: json

            {
                "jsonrpc":"2.0",
                "id":1,
                "result":[
                    100,
                    4,
                ]
            }
        """
        values = [int(v) for v in values]
        return self.modbus_master.execute(int(slave_id),
                                          WRITE_MULTIPLE_COILS,
                                          int(starting_address),
                                          output_value=values)

    @rpcmethod
    @json_rpc_error
    def write_multiple_registers(self, starting_address, values, slave_id=1):
        """ Execute Modbus function code 16: write sequence of values to a
        contiguous block of holding registers.

        .. note:: This method doesn't follow the interface as described in the
            Modbus specification. The specification describes 2 extra
            'parameters': quantity of outputs and byte count. Tolk will derive
            them from parameter `values`.

        :param starting_address: Number of starting address.
        :param values: List with values to write.
        :param slave_id: Number with Slave id, default 1.
        :returns: JSON-RPC response containing the address and the number of values that has been written.

        **Example request:**

        .. sourcecode:: json

            {
                "params":{
                    "starting_address":100,
                    "value": [1234, 5678, 9012, 3456],
                    "slave_id":1
                },
                "jsonrpc":"2.0",
                "method":"write_multiple_registers",
                "id":1
            }

        **Example response:**

        .. sourcecode:: json

            {
                "jsonrpc":"2.0",
                "id":1,
                "result":[
                    100,
                    4,
                ]
            }
        """
        values = [int(v) for v in values]
        return self.modbus_master.execute(int(slave_id),
                                          WRITE_MULTIPLE_REGISTERS,
                                          int(starting_address),
                                          output_value=values)
